import browser.document;
import browser.window;
import jslang.math;

pragma(no_js_output) {
	enum PI = 3.14159265358979323846;
	float degreesToRadians(float x) {
		return (x * PI / 180);

	}

	// Search "user-defined literals in the D programming language"
	// on the web for an explanation of this
	template degrees(float x) {
		enum degrees = degreesToRadians(x);
	}
}

struct Player {
	float x = 640/2;
	float y = 480/2;
	float angle = 0;
	float dx = 0;
	float dy = 0;
	float da = 0;
	bool rotatingLeft;
	bool rotatingRight;
	bool thrusting;

	void advanceFrame() {
		if(rotatingLeft)
			da -= degrees!1;
		if(rotatingRight)
			da += degrees!1;

		angle += da;

		if(thrusting) {
			dx += Math.cos(angle);
			dy += Math.sin(angle);
		}

		x += dx;
		y += dy;

		if(x < 0)
			x += 640;
		if(y < 0)
			y += 480;
		if(x > 640)
			x -= 640;
		if(y > 480)
			y -= 480;
	}

	void draw(JSCanvasContext painter) {
		float size = 15;

		painter.strokeStyle = "#ffffff";

		painter.beginPath();

		painter.moveTo(x, y);

		foreach(a; [degrees!(180-30), degrees!0, degrees!(180+30)]) {
			painter.lineTo( x + size * Math.cos(a + angle), y + size * Math.sin(a + angle));
		}

		painter.lineTo(x, y);

		painter.closePath();
		painter.stroke();
	}
}


JSElement canvas;
Player player;

/*
whoa whoa whoa whoa whoa. add orbits too so the circles themselves move and you have to stay with them.

but anyway so it'd be a 2 minute game. the goal is simply to spend as much time in the circles as you can, but after sitting in one for five seconds, it disappears and randomly moves, so you have to chase it down or move on to another one. in multiplayer you can bump the other people out of their spots.

the smaller the circle, the more points you get.

*/

void frame() {
	player.advanceFrame();

	auto painter = canvas.getContext("2d");
	painter.fillStyle = "#000000";
	painter.fillRect(0, 0, 640, 480); // clear the screen

	painter.strokeStyle = "#33ff33";
	painter.beginPath();
	painter.arc(75, 75, 45, 0, degrees!360, true);
	painter.closePath();
	painter.stroke();

	painter.strokeStyle = "#33ff33";
	painter.beginPath();
	painter.arc(275, 375, 15, 0, degrees!360, true);
	painter.closePath();
	painter.stroke();

	player.draw(painter);
}

void main() {
	canvas = document.getElementById("game");
	auto interval = window.setInterval(&frame, 50);

	document.mainBody.addEventListener("click", {
		// play/pause on click
		if(interval) {
			window.clearInterval(interval);
			interval = 0;
		} else {
			interval = window.setInterval(&frame, 50);
		}
	}, false);

	void keyEventHandler(JSEvent event) {
		auto down = event.type == "keydown";
		switch(event.keyCode) {
			case 'A', 'a', KeyCode.LeftArrow:
				player.rotatingLeft = down;
			break;
			case 'D', 'd', KeyCode.RightArrow:
				player.rotatingRight = down;
			break;
			case 'W', 'w', KeyCode.UpArrow, ' ':
				player.thrusting = down;
			break;
			default: {}
		}

		if(!event.ctrlKey)
			event.preventDefault();
	}

	document.addEventListener("keydown", &keyEventHandler, false);
	document.addEventListener("keyup", &keyEventHandler, false);

	// if("test" in document.__js_this) {}
}
