module arsd.dom;

/* This is a subset of my real dom.d, ported for javascript efficiency */

import jslang.string;
import std.string; // mini phobos is good enough
import browser.document;

extern(js) Document document;

final class Document {
	extern(js) JSDocument __js_this;
	alias __js_this this;

	mixin ElementNavigationFunctions!(Element);
	mixin DocumentNavigationProperties!(Element);
}

final class Element {
	extern(js) JSElement __js_this;
	alias __js_this this;

	mixin ElementNavigationFunctions!(Element);
	mixin ElementNavigationProperties!(Element);

	@property string innerText(string s) {
		innerHtml = "";
		appendChild(document.createTextNode(s));
		return s;
	}

	@property string tagName() {
		// javascript does these uppercase, but dom.d does them lowercase almost always
		auto s = cast(JSString) __js_this.tagName;
		return s.toLowerCase();
	}

	void removeFromTree() {
		if(parentNode)
			parentNode.removeChild(this);
	}

	/// get all the classes on this element
	@property string[] classes() {
		return split(className, " ");
	}

	/// Adds a string to the class attribute. The class attribute is used a lot in CSS.
	Element addClass(string c) {
		if(hasClass(c))
			return this; // don't add it twice

		string cn = getAttribute("class");
		// FIXME: null might need special work here since the explicit check isn't needed in real D
		if(!cn || cn.length == 0) {
			setAttribute("class", c);
			return this;
		} else {
			setAttribute("class", cn ~ " " ~ c);
		}

		return this;
	}

	/// Removes a particular class name.
	Element removeClass(string c) {
		if(!hasClass(c))
			return this;
		string n;
		foreach(name; classes) {
			if(c == name)
				continue; // cut it out
			if(n.length)
				n ~= " ";
			n ~= name;
		}

		className = n.strip;

		return this;
	}

	/// Returns whether the given class appears in this element.
	bool hasClass(string c) {
		auto cn = className;

		auto idx = cn.indexOf(c);
		if(idx == -1)
			return false;

		foreach(cla; cn.split(" "))
			if(cla == c)
				return true;
		return false;
	}

	/*
	// waiting on a compiler bug so alias this works along with opDispatch
	@property string opDispatch(string name)(string value = null) if(name != "popFront" && !__traits(hasMember, typeof(__js_this), name)) {
		if(value is null)
			return getAttribute(name);
		setAttribute(name, value);
		return value;
	}
	*/
}
