module browser.window;

extern(js):
pragma(no_js_output):

import browser.document;
import browser.storage;

final class JSWindow {
	void alert(string);
	bool confirm(string);
	string prompt(string);

	JSWindow open(string url, string name = null, string specs = null, bool replace = false);

	void blur();
	void close();
	void focus();
	void scroll(int, int);
	void scrollBy(int, int);
	void scrollTo(int, int);
	void resizeBy(int, int);
	void resizeTo(int, int);
	void moveBy(int, int);
	void moveTo(int, int);

	void print();

	int setTimeout(void function(), int);
	int setInterval(void function(), int);

	// WARNING: firefox (prior to version 13) would pass an extra parameter to the function you give here.
	int setTimeout(void delegate(), int);
	int setInterval(void delegate(), int);

	void clearTimeout(int);
	void clearInterval(int);

	void function() onload;
	void function() onpopstate;

	JSWindow top;
	JSWindow parent;
	JSWindow opener;
	JSWindow self;

	string status;

	string name;
	bool closed;
	string defaultStatus;

	int innerHeight;
	int innerWidth;
	int outerWidth;
	int outerHeight;
	int pageXOffset;
	int pageYOffset;

	int screenLeft;
	int screenTop;
	int screenX;
	int screenY;

	JSScreen screen;

	// these tend to be available elsewhere

	JSDocument document;
	JSEvent event;

	JSStorage localStorage;
	JSStorage sessionStorage;

	JSLocation location;
	JSHistory history;
}

final class JSScreen {
	int availHeight;
	int availWidth;
	int colorDepth;
	int height;
	int pixelDepth;
	int width;
}

final class JSLocation {
	string hash;
	string host;
	string hostname;
	string href;
	string pathname;
	int port;
	string protocol;
	string search;

	void assign(string);
	void reload();
	void replace(string);
}

final class JSHistory {
	int length;
	void back();
	void forward();
	void go(int);
	void go(string);

	Object state;
	void pushState(Object state, string title, string url);
}

JSWindow window;
JSLocation location;
JSHistory history;
