// FIXME: implement dynamic type that meshes with JS itself well enough
// FIXME: all user defined classes need constructors, but all JS defined
// classes must not have them....

/*
	Changes to the compiler:

	extern(js) are unmangled, can be variadic with or without parameters. These try to be native in calling convention

	pragma(no_js_output) means don't write *any* of this out. You can't call them, including construction, and dynamic
	casts won't work from D (this is not checked btw); use it just for pure wrappers around native stuff to trim useless rtti.
*/

alias immutable(char)[] string;
// the next few are really only there for phobos... they don't actually work right
alias immutable(wchar)[] wstring;
alias immutable(dchar)[] dstring;
alias uint size_t;
alias uint sizediff_t;
alias int ptrdiff_t;
import jslang.global;
alias String text;

/* ******************************** */
/*          Basic D classes         */
/* ******************************** */

// the D language expects these to be around. Let's try to match JS as well as we can.
class Object {
	extern(js) {
		extern(js) Object[string] __js_this; // a compiler supported hack to give access to more native functions

		Object prototype; /// the prototype object from javascript
		string delegate() __js_toString; /// for Javascript overriding

		/// these are filled in by the D compiler
		string[] __d_implements; // this is important for dynamic casting. __d_implements[0] == this.mangleof
		string __d_classname; // can use it for pretty printing
	}

	string toString() { return ""; } // for D
	bool opEquals(Object rhs) { return rhs is this; }
}
class Throwable : Object { // required by the D compiler
	string message;
	alias message msg;
	string file;
	int line;
	this(string msg, string file = __FILE__, int line = __LINE__) {
		this.message = msg;
		this.file = file;
		this.line = line;
		super.__js_toString = &toString;
	}
	override string toString() {
		return  file ~ ":" ~ String(line) ~ " [" ~ __d_classname ~ "] " ~ message;
	}
}
class Error : Throwable { // required by the D compiler
	this(string msg, string file = __FILE__, int line = __LINE__) {
		super(msg, file, line);
	}
}
class Exception : Throwable { // required by the D compiler
	this(string msg, string file = __FILE__, int line = __LINE__) {
		super(msg, file, line);
	}
}

/// thrown by our assert function
class AssertError : Error {
	this(string msg, string file = __FILE__, int line = __LINE__) {
		super(msg, file, line);
	}
}



extern(js) {

	/* ******************************** */
	/*   Access to Javascript language  */
	/* ******************************** */

	// access to JS keywords. the type isn't a perfect match
	Object __js_this; // this special thing translates directly to the JS keyword "this". May not refer to what you think it does, even if you're familiar with javascript, due to how the D compiler is free to arrange things
	Object[] __js_arguments; // the arguments keyword in JS. ditto on warning.

	// we need some definitions from the js language...
	import jslang.array;
	import jslang.global;
	import jslang.func;

	// initializer helpers
	Array __js_array_literal; // outputs as []
	Object __js_object_literal; // outputs as {}


	/* ******************************** */
	/*        D runtime functions       */
	/* ******************************** */

	// called if asserts fail
	void __d_assert_msg(string msg, string file, int line) {
		throw new AssertError(msg, file, line);
	}
	void __d_assert(string file, int line) {
		throw new AssertError("Assertion failure", file, line);
	}

	// given an object, should we allow the runtime cast to succeed?
	// if no, return null.
	Object __d_dynamic_cast(Object from, string to) {
		if(from.__d_implements) {
			foreach(s; from.__d_implements)
				if(s == to)
					return from;
			return null;
		}
		// if in doubt, let it pass since we don't want to tie hands

		return from;
	}

	// if you do [1,2,3] == [1,2,3], it calls this function
	bool __d_array_compare(in void[] a, in void[] b) {
		if(a.length != b.length)
			return false;
		// it isn't necessarily floats, but it is good enough for our use case; this is as native as you get in JS for numeric element comparisons
		auto bytes = cast(float[]) a;
		auto bytes2 = cast(float[]) b;
		for(int i = 0; i < bytes.length; i++)
			if(bytes[i] != bytes2[i])
				return false;
		return true;
	}

	/// This is called after the initializer, in the JS constructor function, with JS's this set to the new object. It is passed a D constructor as the first argument, and the arguments array from the new function as the second argument (remember, the constructor is #1 there)
	Object __d_construct(Function constructor, Object[] arguments) {
		if(!constructor) return __js_this;

		Array constructorArguments = __js_array_literal;
		for(int a = 1; a < arguments.length; a++)
			constructorArguments.push(arguments[a]);
		return constructor.apply(__js_this, constructorArguments);
	}

	// note: _aaApply and _aaApply2 are implemented as functions in the compiler since I can't do them efficiently here...
	// the D compiler needs this to be here, though we do nothing with it
	pragma(no_js_output) {
		struct AssociativeArray(K, V) {
			typeof(this) dup() { return this; } // FIXME
		}
		class TypeInfo {}
		class TypeInfo_Class : TypeInfo {}
		class TypeInfo_Array : TypeInfo {}
		class ClassInfo {}
	}

	// FIXME: this is actually kinda broken
	// the idea though is to just forcibly convert the type in D, while being a no-op in JS; it should show as nothing more than a plain assignment there.
	pragma(no_js_output)
	T reinterpret_cast(T, R)(R r) {
		return cast(T) cast(void*) r;
	}

	// dmdscript functions
	void println(...);//string s);
}

// these give a little more access to the native environment
// ideallly, it won't need wrapper functions anymore, but for now it is easier to use
bool hasProperty(Object o, string propertyName) {
	return (propertyName in o.__js_this) ? true : false;
}
