// this strips out functions that are never used
// call this BEFORE calling mangledown

// it is especially useful if you called dmd with -inline

import std.string;
import std.array;
import std.file;
import std.regex;

import std.stdio;
import std.conv;

string doReplaceFunctions(string s) {
	startover:
	foreach(m; s.match(regex("function [a-zA-Z0-9_]+", "g"))) {
		auto full = to!string(m[0]);
		auto name = std.array.replace(full, "function ", "");

		int count = 0;
		int definitions = 0;
		foreach(call; s.match(regex("\\b" ~ name ~ "\\b", "g")))
			if(!endsWith(call.pre, "function ")) // the declaration doesn't count
				count++;
			else {
				definitions++;
			}

		if(definitions > 1) {
			writeln("multiple definitions of ", full);
		}

		if(count == 0 || definitions > 1) {
			// it isn't called or referenced, let's kill it

			auto idx = s.indexOf(full);
			assert(idx != -1);

			auto end = idx;
			bool started = false;
			int braceCount = 0;
			char instr = 0;
			bool escaping = false;
			foreach(ch; s[idx .. $]) {
				end++;

				if(instr != 0) {
					if(escaping) {
						escaping = false;
						continue;
					}

					if(ch == '\\')
						escaping = true;
					else
						escaping = false;

					if(ch == instr)
						instr = false;

					continue;
				}

				if(ch == '"' || ch == '\'')
					instr = ch;
				if(ch == '{') {
					started = true;
					braceCount++;
				}

				if(ch == '}')
					braceCount--;

				if(started && braceCount == 0)
					break;
			}

			auto news = s[0 .. idx];
			assert(end <= s.length, to!string(end) ~ " / " ~ to!string(s.length));
			if(end < s.length && s[end] == '\n')
				end++; // trim blank lines too while we're at it
			news ~= s[end .. $];

			s = news;

			writeln(full, " was garbage");

			goto startover;
		}
	}

	return s;
}

bool hasPossibleSideEffect(string code) {
	char instr = 0;
	bool escaping = false;
	foreach(ch; code) {
		if(instr != 0) {
			if(escaping) {
				escaping = false;
				continue;
			}

			if(ch == '\\')
				escaping = true;
			else
				escaping = false;

			if(ch == instr)
				instr = false;

			continue;
		}

		if(ch == '"' || ch == '\'')
			instr = ch;

		if(ch == ';')
			return false; // end of statement

		if(ch == '(') // we'll assume all parens might be function calls
			return true;

		assert(ch != '\n');
	}

	return false;
}

string doReplaceVars(string s) {
	startover:
	foreach(m; s.match(regex("var [a-zA-Z0-9_]+", "g"))) {
		auto full = to!string(m[0]);
		auto name = std.array.replace(full, "var ", "");

		int count = 0;
		foreach(call; s.match(regex("\\b" ~ name ~ "\\b", "g"))) {
			// if it is just another variable declaration, it isn't actually a use, unless there's a function call in there.
			if(!endsWith(call.pre, "var ") || hasPossibleSideEffect(call.post))
				count++;
		}

		if(count == 0) {
			// it isn't referenced, let's kill it

			auto idx = s.indexOf(full);
			assert(idx != -1);

			auto end = idx;
			char instr = 0;
			bool escaping = false;
			foreach(ch; s[idx .. $]) {
				end++;

				if(instr != 0) {
					if(escaping) {
						escaping = false;
						continue;
					}

					if(ch == '\\')
						escaping = true;
					else
						escaping = false;

					if(ch == instr)
						instr = false;

					continue;
				}

				if(ch == '"' || ch == '\'')
					instr = ch;

				if(ch == ';')
					break;
				assert(ch != '\n');
			}

			auto news = s[0 .. idx];
			assert(end <= s.length, to!string(end) ~ " / " ~ to!string(s.length));
			if(end < s.length && s[end] == '\n')
				end++; // trim blank lines too while we're at it
			news ~= s[end .. $];

			s = news;

			writeln(full, " was garbage");

			goto startover;
		}
	}

	return s;
}


void main(string[] args) {
	auto src = readText(args[1]);

	src = src.doReplaceFunctions();
	src = src.doReplaceVars();
	// second pass, just in case
	src = src.doReplaceFunctions();
	src = src.doReplaceVars();

	std.file.write(args[1], src);
}
