import browser.document;
import browser.window;

class Game {
	// implement these two functions
	/// This function is called once per interval. You should update your game state
	abstract void update();
	/// This is called at some point after update, when possible. You should draw your scene.
	abstract void draw();
	/// The default key handler keeps this set up so you can check key state
	bool[124] keyState;

	int fps;
	/// fps is the number of updates and the max fps it will draw.
	this(int fps, JSElement holder = null) {
		this.fps = fps;
		if(holder is null)
			holder = document.mainBody;
		holder.addEventListener("keydown", &keyEventHandler, false);
		holder.addEventListener("keyup", &keyEventHandler, false);
	}

	void keyEventHandler(JSEvent event) {
		auto down = event.type == "keydown";
		if(event.which < keyState.length)
			keyState[event.which] = down;
		event.preventDefault();
	}

	bool needsDrawing;
	int drawTimer;
	void frameEvent() {
		if(drawTimer)
			window.clearTimeout(drawTimer);
		needsDrawing = true; // assume that an update needs drawing
		update();

		if(needsDrawing && !drawTimer)
			drawTimer = window.setTimeout(&drawFrameEvent, 1); // draw asap, but don't block this thread
	}

	void drawFrameEvent() {
		if(!needsDrawing)
			return;
		draw();
		needsDrawing = false;
		drawTimer = 0;
	}

	int frameInterval;
	void play() {
		frameInterval = window.setInterval(&frameEvent, 1000 / fps);
	}
}

struct Particle {
	int id;
	float x;
	float y;
	float dx;
	float dy;
}

import std.conv;

class Test : Game {
	int width = 320;
	int height = 240;

	this() {
		foreach(i; 0..terrain.length)
			if(i < 320*120)
				terrain[i] = 0;
			else
				terrain[i] = 1;

		colors = ["#ccccff", "#33ff33"];

		canvas = document.getElementById("game");
		canvas.setAttribute("width", to!string(width));
		canvas.setAttribute("height", to!string(height));
		draw();

		super(2);
	}

	int terrain[320*240] = void;
	string[] colors;
	Particle[] particlesInMotion;
	JSElement canvas;

	override void update() {
		
		needsDrawing = false; // the browser does this
	}

	override void draw() {
		auto painter = canvas.getContext("2d");

		int idx = 0;
		for(int y = 0; y < 240; y++) {
		for(int x = 0; x < 320; x++) {
			painter.fillStyle = colors[terrain[idx]];
			painter.fillRect(x, y, x+1, y+1);
			idx++;
		}
		}
		return;
	}
}

void main() {
	auto game = new Test();
	game.play();
}
