import browser.xmlhttp;
import browser.document;
import browser.json;

import jslang.global;
import jslang.array;

extern(js)
final class Envelope {
	bool success;
	string type;
	string errorMessage;
	string userData;

	Object result;
	string dFullString;
}

void webDotDRequest(string url, void delegate(Object) success, void delegate(string) error= null) {
	auto http = new XMLHttpRequest();
	http.open("GET", url, true);

	http.onreadystatechange = {
		if(http.readyState == 4) {
			if(http.getAllResponseHeaders()) {
				auto obj = cast(Envelope) JSON.parse(http.responseText);
				if(obj.success) {
					success(obj.result);
				} else if(error) {
					error(obj.errorMessage);
				} else {
					throw new Exception(obj.errorMessage);
				}
			}
		}
	};

	http.send();
}

extern(js)
final class GraphData {
	string[] x;
	int[] y;
}

void doGraph(JSElement canvas, GraphData data) {
	auto painter = canvas.getContext("2d");
	int width = Number(canvas.getAttribute("width"));
	int height = Number(canvas.getAttribute("height"));

	painter.fillStyle = "rgba(200, 220, 255, 0.9)";

	// for our text
	painter.strokeStyle = "black";
	painter.lineWidth = 0.5;

	painter.beginPath();

	auto drawHeight = height - 15;
	auto drawOffset = 30;
	auto drawWidth = width - drawOffset;
	painter.moveTo(drawOffset, drawHeight - 1); // open our graph

	int min = int.max, max = int.min;

	Array ySorted = __js_array_literal;
	foreach(y; data.y) {
		if(y < min)
			min = y;
		if(y > max)
			max = y;
		ySorted.push(y);
	}

	auto range = max - min;

	int lastX;
	int lastXUsed;

	foreach(i, x; data.x) {
		auto y = data.y[i];

		lastX = cast(int) (drawWidth * cast(float) i / data.x.length) + drawOffset;

		painter.lineTo(
			lastX,
			drawHeight - (drawHeight * cast(float) (y-min) / range)
		);

		if(i == 0 || lastX - lastXUsed > 25) {
			painter.strokeText(x, lastX, height - 1);
			lastXUsed = lastX;
		}
	}

	painter.lineTo(lastX, drawHeight - 1);
	painter.lineTo(drawOffset + 2, drawHeight - 1); // closing our graph

	painter.closePath();
	painter.fill();

	painter.strokeStyle = "#1d73ac";
	painter.lineWidth = 2;
	painter.stroke();

	// for our text
	painter.strokeStyle = "black";
	painter.lineWidth = 0.5;

	ySorted.sort(function(int a, int b) { return b-a; });

	int lastYUsed;
	int ypos;
	foreach(i, y; cast(int[]) ySorted) {
		ypos = drawHeight - (cast(int)(drawHeight * (cast(float) (y - min) / range))) + 10;
		if(i == 0 || ypos - lastYUsed > 8) {
			painter.strokeText(String(y), 0, ypos);
			lastYUsed = ypos;
		}
	}

}

/// public
void graphTo(JSElement canvas, string dataUrl) {
	webDotDRequest(dataUrl, (Object o) {
		doGraph(canvas, cast(GraphData) o);
	});
}
